# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from collections import Counter
from blueqat.backends.qasm_parser_backend_generator import generate_backend
import re

def draw_circuit_obaq(circuit, nqubits, depth):
  import networkx as nx
  import matplotlib.pyplot as plt

  nqubits = nqubits
  depth = depth

  width = 11

  #for lines
  G = nx.Graph()

  #for init qubit
  Gi = nx.Graph()

  #just for label
  GCX_l = nx.Graph()
  GCX_l.add_node((0,0))

  #for CX
  GCX = nx.Graph()

  #for Controlled bit
  GC = nx.Graph()

  #for pauli&H
  GX = nx.Graph()
  GY = nx.Graph()
  GZ = nx.Graph()
  GH = nx.Graph()

  #initial qubits and measurement
  G.add_nodes_from([(0, i) for i in range(nqubits)])
  G.add_nodes_from([(depth+3, i) for i in range(nqubits)])

  #initial qubits again
  Gi.add_nodes_from([(0, i) for i in range(nqubits)])

  #edge from start to final
  for i in range(nqubits):
    G.add_edge((0, i),(depth+3, i))

  #for CX
  gcx = circuit['cx']

  for i in range(len(gcx)):
    GCX.add_node((gcx[i][0], gcx[i][1]))
    GCX.add_node((gcx[i][0], gcx[i][2]))
    GC.add_node((gcx[i][0], gcx[i][1]))
    GCX.add_edge((gcx[i][0], gcx[i][1]), (gcx[i][0], gcx[i][2]))

  #for X
  GX.add_nodes_from(circuit['x'])

  #for Y
  GY.add_nodes_from(circuit['y'])

  #for Z
  GZ.add_nodes_from(circuit['z'])

  #for H
  GH.add_nodes_from(circuit['h'])

  #color
  color_cx = "#04C4D9"

  nx.draw(G,
        pos=dict((n, (n[0], -n[1])) for n in G.nodes()),
        node_color = 'white',
        edge_color = 'lightgrey',
        width = 1.5,
        node_size=100)

  nx.draw(GCX_l,
        node_color = color_cx,
        label = "CX",
        node_size=100)

  nx.draw(Gi,
        pos=dict((n, (n[0], -n[1])) for n in Gi.nodes()),
        node_color = 'grey',
        node_size=100)

  nx.draw(GCX,
        pos=dict((n, (n[0]+1, -n[1])) for n in GCX.nodes()),
        node_color = color_cx,
        edge_color = color_cx,
        width = width,
        node_size=100)

  #controlled bit
  nx.draw(GC,
        pos=dict((n, (n[0]+1, -n[1])) for n in GC.nodes()),
        node_color = 'white',
        edge_color = 'gray',
        width = 1.5,
        node_size=30)

  #X
  nx.draw(GX,
        pos=dict((n, (n[0]+1, -n[1])) for n in GX.nodes()),
        node_color = '#F2B705',
        label = "X",
        node_size=100)

  #Y
  nx.draw(GY,
        pos=dict((n, (n[0]+1, -n[1])) for n in GY.nodes()),
        node_color = '#F28705',
        label = "Y",
        node_size=100)
  
  #Z
  nx.draw(GZ,
        pos=dict((n, (n[0]+1, -n[1])) for n in GZ.nodes()),
        node_color = '#F25C05',
        label = "Z",
        node_size=100)

  #H
  nx.draw(GH,
        pos=dict((n, (n[0]+1, -n[1])) for n in GH.nodes()),
        node_color = '#F21313',
        label = "H",
        node_size=100)

  plt.legend(bbox_to_anchor=(1.15, 1),loc='upper right')
  plt.axis('equal')
  plt.show()

def _qasm_runner_obaq(qasm, shots=None, returns=None, **kwargs):
    if returns is None:
        returns = "shots"
    elif returns not in ("shots", "draw", "_exception"):
        raise ValueError("`returns` shall be None, 'shots', 'draw', " +
                         "'_exception'")

    import_error = None
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
    except Exception as e:
        import_error = e

    if import_error:
        if returns == "_exception":
            return e
        if isinstance(import_error, ImportError):
            raise ImportError("Cannot import qiskit. To use this backend, please install networkx." +
                              " `pip install networkx `.")
        else:
            raise ValueError("Unknown error raised when importing networkx. To get exception, " +
                             'run this backend with arg `returns="_exception"`')
    else:
        if returns == "_exception":
            return None

        #parse qasm
        circ = qasm.split(';\n')
        del circ[:2]

        #nqubits
        nqubits = int(re.compile('\d+').findall(circ[0].split(' ')[1])[0])
        circ = [i for i in circ if i[:7] != "measure" and i[:4] != "qreg" and i[:4] != "creg"]

        #depth
        depth = len(circ)

        #count of gates
        circ2 = {'x':[], 'y':[], 'z':[], 'h':[], 'cx':[], 'cy':[], 'cz':[], 'rx':[], 'ry':[], 'rz':[]}
        for i in range(depth):
          temp = circ[i].split(' ')
          if temp[0] == 'cx':
            pos1 = int(re.compile('\d+').findall(temp[1].split(',')[0])[0])
            pos2 = int(re.compile('\d+').findall(temp[1].split(',')[1])[0])
            circ2[temp[0]].append((i, pos1, pos2))
          else:
            circ2[temp[0]].append((i, int(re.compile('\d+').findall(temp[1])[0])))
 
        #print(circ2)

        if returns == "draw":
          return draw_circuit_obaq(circ2, nqubits, depth)
        if shots is None:
          shots = 1024

        return circ2

obaq_backend = generate_backend(_qasm_runner_obaq)

from blueqat import BlueqatGlobalSetting
BlueqatGlobalSetting.register_backend("obaq", obaq_backend)
